use std::fs;

const DECRYPTION_KEY: i64 = 811_589_153;

pub fn solve() {
    let path = "res/20/input.txt";

    let contents = fs::read_to_string(path).expect("I/O error, wrong path?");

    let initial_arrangement: Vec<i64> = contents
        .lines()
        .filter(|line| *line != "")
        .map(|line| line.parse::<i64>().unwrap())
        .collect();

    let size = initial_arrangement.len();

    // list of indices
    let mut work_arrangement: Vec<usize> = initial_arrangement
        .iter()
        .enumerate()
        .map(|(i, _)| i)
        .collect();

    for (i, &n) in initial_arrangement.iter().enumerate() {
        // position in current state
        let work_i = work_arrangement.iter().position(|&elem| elem == i).unwrap();
        mv(work_i, n, &mut work_arrangement);
        //println!("{}\n", work_arrangement.iter().map(|i| initial_arrangement[*i].to_string()).collect::<Vec::<String>>().join(", "));
    }

    let offset = work_arrangement
        .iter()
        .position(|i| initial_arrangement[*i] == 0)
        .unwrap();
    let coord_1 = initial_arrangement[work_arrangement[(1000 + offset) % size]];
    let coord_2 = initial_arrangement[work_arrangement[(2000 + offset) % size]];
    let coord_3 = initial_arrangement[work_arrangement[(3000 + offset) % size]];

    println!("Result 1: {}", coord_1 + coord_2 + coord_3);

    // task 2
    let decrypted_arrangement: Vec<i64> = initial_arrangement
        .iter()
        .map(|i| i * DECRYPTION_KEY)
        .collect();

    // list of indices
    let mut work_arrangement: Vec<usize> = decrypted_arrangement
        .iter()
        .enumerate()
        .map(|(i, _)| i)
        .collect();

    // do the whole thing 10x
    for _ in 0..10 {
        for (i, &n) in decrypted_arrangement.iter().enumerate() {
            // position in current state
            let work_i = work_arrangement.iter().position(|&elem| elem == i).unwrap();
            mv(work_i, n, &mut work_arrangement);
            //println!("{}\n", work_arrangement.iter().map(|i| decrypted_arrangement[*i].to_string()).collect::<Vec::<String>>().join(", "));
        }
    }

    let offset = work_arrangement
        .iter()
        .position(|i| initial_arrangement[*i] == 0)
        .unwrap();
    let coord_1 = decrypted_arrangement[work_arrangement[(1000 + offset) % size]];
    let coord_2 = decrypted_arrangement[work_arrangement[(2000 + offset) % size]];
    let coord_3 = decrypted_arrangement[work_arrangement[(3000 + offset) % size]];

    println!("Result 2: {}", coord_1 + coord_2 + coord_3);
}

fn mv(pos: usize, move_by: i64, arr: &mut Vec<usize>) {
    // reduce move_by to modulo
    let move_by = if move_by < 0 {
        (move_by.abs() % (arr.len() as i64 - 1)) * -1
    } else {
        move_by % (arr.len() as i64 - 1)
    };

    let move_by = if pos as i64 + move_by <= 0 {
        move_by + arr.len() as i64 - 1
    } else if pos as i64 + move_by >= arr.len() as i64 - 1 {
        move_by - arr.len() as i64 + 1
    } else {
        move_by
    };

    let mut curr_pos = pos;
    let change = if move_by < 0 { -1 } else { 1 };

    for _ in 0..move_by.abs() {
        let other_pos = ((curr_pos + arr.len()) as i64 + change) as usize % arr.len();
        let val = arr[curr_pos];
        arr[curr_pos] = arr[other_pos];
        arr[other_pos] = val;
        curr_pos = other_pos;
    }
}
